use std::{collections::BTreeMap, path::PathBuf};

use anyhow::anyhow;
use clap::{Parser, Subcommand};

#[derive(Parser)]
#[command(author, version, about, long_about = None, subcommand_required = true)]
struct Command {
    #[command(subcommand)]
    command: Option<Commands>,
}

#[derive(Subcommand)]
enum Commands {
    Upload {
        #[arg(long = "template")]
        template: String,

        #[arg(long = "drone-token", env = "DRONE_TOKEN")]
        drone_token: String,

        #[arg(long = "drone-host", env = "DRONE_HOST")]
        drone_host: String,

        #[arg(long = "drone-user", env = "DRONE_USER")]
        drone_user: String,

        #[arg(long = "variable", short = 'v')]
        variables: Vec<String>,
    },
}

#[tokio::main]
async fn main() -> anyhow::Result<()> {
    dotenv::dotenv().ok();
    tracing_subscriber::fmt::init();

    let cli = Command::parse();
    tracing::debug!("starting cli");

    if let Some(Commands::Upload {
        template,
        drone_token,
        drone_host,
        drone_user,
        variables,
    }) = cli.command
    {
        tracing::info!(template, drone_host, "executing drone upload");

        let vars = variables
            .iter()
            .filter_map(|v| v.split_once('='))
            .collect::<BTreeMap<&str, &str>>();

        let template_path = PathBuf::from(&template);
        if !template_path.exists() {
            anyhow::bail!("template: {} doesn't exist", template)
        }

        let template_name = template_path
            .file_name()
            .ok_or(anyhow!(
                "template: {} didn't include a proper filename",
                &template
            ))?
            .to_str()
            .expect("filename to be utf8");
        let template_data = tokio::fs::read_to_string(&template_path).await?;

        let mut ctx = tera::Context::new();
        for (name, val) in vars {
            tracing::debug!("adding value to template: ({}, {})", &name, &val);
            ctx.insert(name, &val);
        }
        let template_data = tera::Tera::one_off(&template_data, &ctx, false)?;

        let template_url = format!(
            "{}/api/templates/{}/{}",
            drone_host.trim_end_matches('/'),
            drone_user,
            template_name
        );

        let client = reqwest::Client::new();

        match client
            .delete(&template_url)
            .bearer_auth(&drone_token)
            .send()
            .await
        {
            Ok(res) => match res.error_for_status() {
                Ok(_) => {}
                Err(status) => {
                    tracing::warn!("failed to delete template {}", status);
                }
            },
            Err(e) => {
                tracing::error!("{}", e);
            }
        }

        let res = client
            .put(&template_url)
            .bearer_auth(&drone_token)
            .json(&serde_json::json!({
                "data": &template_data,
            }))
            .send()
            .await?;

        if let Err(e) = res.error_for_status() {
            match e.status() {
                Some(reqwest::StatusCode::NOT_FOUND) => {
                    tracing::debug!("template was not found attemping creation instead");

                    let template_url = format!(
                        "{}/api/templates/{}",
                        drone_host.trim_end_matches('/'),
                        drone_user,
                    );

                    let res = client
                        .post(&template_url)
                        .bearer_auth(&drone_token)
                        .json(&serde_json::json!({
                            "name": &template_name,
                            "data": &template_data,
                        }))
                        .send()
                        .await?;

                    if let Err(e) = res.error_for_status() {
                        anyhow::bail!(e)
                    }

                    tracing::debug!("template creation successful");
                }
                Some(status) => {
                    anyhow::bail!("error: {}, failed to upload template: {}", e, status)
                }
                None => {
                    anyhow::bail!(e)
                }
            }
        }

        tracing::info!(template, "successfully updated template");
    }

    Ok(())
}
